#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import math
import numpy as np
import scipy.spatial

import collections
import heapq

import Environment as env
import SamplingPlanner as sp

import dubins

class RRTPlanner(sp.SamplingPlanner):

    def __init__(self, limits, turning_radius):
        self.turning_radius = turning_radius
        sp.SamplingPlanner.__init__(self,limits)    

    def create_dubins_path(self, start, end):

        # get euler angles
        te1 = sp.se3_2_te(start)
        te2 = sp.se3_2_te(end)

        # create dubins states
        q1 = [te1[0],te1[1],te1[5]]
        q2 = [te2[0],te2[1],te2[5]]

        # compute shortest dubins path
        pth = dubins.shortest_path(q1, q2, self.turning_radius)

        return pth

    def check_collision(self, pth, environment, verbose = False):
        ret = False

        for se3 in pth:
            ret = (se3[0,3] < self.limits[0][0] or se3[0,3] > self.limits[0][1] or se3[1,3] < self.limits[1][0] or se3[1,3] > self.limits[1][1])
            if ret:
                if verbose:
                    print ('out of bounds at %s' % str(se3))
                break
            ret = environment.check_robot_collision(se3)
            if ret:
                if verbose:
                    print ('collision at %s' % str(se3))
                break
        return ret   

    def create_sub_path(self, start, end, dubins_pth = None):

        if dubins_pth is None:
            dubins_pth = self.create_dubins_path(start, end)

        sub_pth = []
        for qq in dubins_pth.sample_many(self.collision_resolution_translation)[0]: 

            te = [qq[0],qq[1],0,0,0,qq[2]]
            se3 = sp.te_2_se3(te)
            sub_pth.append(se3)

        return sub_pth    


    def is_valid_edge(self, start, end, environment, verbose=False, dubins_pth = None):
        """
        Is and oriented edge between two configurations valid? 

        Parameters
        ----------
        start: numpy array (4x4)
            start configuration of the robot in SE(3) coordinates
        end: numpy array (4x4)
            end configuration of the robot in SE(3) coordinates
        environment: Environment
            Map of the environment that provides collision checking 
        verbose: boolean
             if True then verbose

        Returns
        -------
        boolean
            True if the edge is valid
        """

        if dubins_pth is None:
            dubins_pth = self.create_dubins_path(start, end)
 
        ret = True

        if verbose:
            t1 = start[:3,3]
            t2 = end[:3,3]
            print ('te1 : %s' % str(t1))
            print ('te2 : %s' % str(t2))
            print ('dubins len : %.6f' % dubins_pth.path_length())
            print ('max norm : %.6f' % self.max_norm_translation)

        if dubins_pth.path_length() > self.max_norm_translation:
            ret = False
            if verbose:
                print ('fail distance')
        elif self.check_collision(self.create_sub_path(start,end,dubins_pth),environment):
            ret = False
            if verbose:
                print ('fail collision')

        return ret

    def plan(self, environment, start, goal):
        """
        Method to plan the path

        Parameters
        ----------
        environment: Environment
            Map of the environment that provides collision checking 
        start: numpy array (4x4)
            start configuration of the robot in SE(3) coordinates
        goal: numpy array (4x4)
            goal configuration of the robot in SE(3) coordinates

        Returns
        -------
        list(numpy array (4x4))
            the path between the start and the goal Pose in SE(3) coordinates
        """

        #TODO: t3b-rrt - implement the curvature constrained RRT planner

        print (start)
        print (goal)

        assert not environment.check_robot_collision(start), 'start collision'
        assert not environment.check_robot_collision(goal), 'goal collision'

        path = [start,goal]

         
        return(path)
    
